unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls,
  Grids, SpinEx, fpjson, jsonparser;

const
  colID = 1;
  colCty = 2;
  colLon = 4;
  colLat = 3;
  colName = 5;

type
  TCity = record
    ctLat, ctLon: double;
    ctKoorIdx, ctId: longint;
    ctCty: string[3];
    ctName: string[83];
  end;
  tpCity = ^TCity;
  TResIndices = array of longword;
  TPointerList = array of Pointer;
  TCompFunc = function(A, B: Pointer): NativeInt;
  TCompArr = array of TCompFunc;
  { TForm1 }

  TForm1 = class(TForm)
    Button1: TButton;
    SucheKoor: TButton;
    SucheLand: TButton;
    SucheOrt: TButton;
    edName: TEdit;
    edCountry: TEdit;
    Longitude: TFloatSpinEditEx;
    Latitude: TFloatSpinEditEx;
    Label1: TLabel;
    Label2: TLabel;
    Memo1: TMemo;
    StringGrid1: TStringGrid;
    procedure Button1Click(Sender: TObject);
    procedure FormShow(Sender: TObject);

    procedure SucheKoorClick(Sender: TObject);
    procedure SucheLandClick(Sender: TObject);
    procedure SucheOrtClick(Sender: TObject);

  private
    POIs: array of TCity;
    pPOIs: TPointerList;
    ResIdx: TResIndices;
    ResCnt: NativeInt;
    procedure ReadPOIs(OutName: string);
    procedure ConvertJson(InName, OutName: string);
    procedure OutPOI(POIidx: NativeInt);
    procedure CheckCountry(s: string);
    procedure CheckName(s: string);
    procedure CheckCoor(coorIdx: NativeInt);
    procedure FillStringGrid;
  public

  end;

var
  Form1: TForm1;
  CompArr: TCompArr;

implementation

const
// Schreibt in dem aktuellen Verzeichnis
//  f := TFileStream.Create(ExtractFilePath(ParamStr(0)) + OutName
  cInName = 'city.list.json';
  cOutName = 'city.dat';

{$R *.lfm}
{sortieren }

function CompKoorIdxID(A, B: Pointer): NativeInt;
begin
  Result := tpCity(A)^.ctKoorIdx - tpCity(B)^.ctKoorIdx;
end;

function CompID(A, B: Pointer): NativeInt;
begin
  Result := tpCity(A)^.ctID - tpCity(B)^.ctID;
end;

function CompName(A, B: Pointer): NativeInt;
begin
  Result := Ord(tpCity(A)^.ctName >= tpCity(B)^.ctName) -
    Ord(tpCity(A)^.ctName <= tpCity(B)^.ctName);
end;

function CompCty(A, B: Pointer): NativeInt;
begin
  Result := Ord(tpCity(A)^.ctCty >= tpCity(B)^.ctCty) -
    Ord(tpCity(A)^.ctCty <= tpCity(B)^.ctCty);
end;

function CompSeq(A, B: Pointer): NativeInt;
var
  i: NativeInt;
begin
  Result := 0;
  for i := 0 to High(CompArr) do
  begin
    Result := CompArr[i](A, B);
    if Result <> 0 then
      break;
  end;
end;

procedure mergesort(a: TPointerList; CompFunc: TCompFunc);
var
  B: TPointerlist;

  procedure merge(links, rechts: NativeInt);
  var
    pLinks: ^Pointer;
    i, j, k, mid: NativeInt;
  begin
    if links < rechts then
    begin
      // Mittleres Element bestimmen und rekursiver Abstieg
      mid := (rechts + links) div 2;
      merge(links, mid);
      merge(mid + 1, rechts);
      //schon sortiert?
      if CompFunc(A[Mid], A[Mid + 1]) <= 0 then
        EXIT;
      // Mischen der sortierten Unterfelder
      // untere Hälfte der Daten in Hilfsfeld kopieren
      move(A[links], B[0], (mid - links + 1) * SizeOf(Pointer));
      i := 0;
      j := mid + 1;
      k := links;
      pLinks := @A[k];
      //einsortieren der beiden Hälften
      while (k < j) and (j <= Rechts) do
      begin
        if CompFunc(B[i], A[j]) <= 0 then
        begin
          plinks^ := B[i];
          Inc(i);
        end
        else
        begin
          pLinks^ := A[j];
          Inc(j);
        end;
        Inc(k);
        Inc(pLinks);
      end;
      //    Rest in B nach A kopieren
      //    move(B[i],A[k],(j-k)*SizeOf(Pointer));
      while (k < j) do
      begin
        pLinks^ := B[i];
        Inc(pLinks);
        Inc(k);
        Inc(i);
      end;
    end;   // Ende von "if" (Abbruchbedinung)
  end;     // Ende von "procedure"

begin
  setlength(B, (High(A) + 1) div 2 + 1);
  merge(0, High(A));
  setlength(B, 0);
end;

{ TForm1 }
function CalcKoorIdx(Lat, Lon: double): NativeInt; inline;
begin
  Result := trunc(lon + 180) * 256 + trunc(Lat + 90);
end;

procedure TFOrm1.ConvertJson(InName, OutName: string);
var
  f: TFileStream;
  stream: TFileStream;
  json: TJSONData;
  parser: TJSONParser;
  jArr: TJSONArray;
  jcoord: TJSONObject;
  t: TDateTime;
  i, j: integer;
begin
  t := now;
  json := nil;
  try
    stream := TFileStream.Create(ExtractFilePath(ParamStr(0)) + InName,
      fmOpenRead or fmShareDenyWrite);
    try
      parser := TJSONParser.Create(Stream, []);
      try
        json := parser.Parse;
      finally
        parser.Free;
      end;
    finally
      stream.Free;
    end;
    j := json.Count;
    SetLength(POIs, j);
    SetLength(pPOIs, j);

    jArr := TJSONArray(json);
    for j := 0 to jArr.Count - 1 do
    begin
      with POIs[j], TJSONObject(jArr[j]) do
      begin
        ctID := Find('id').AsInteger;
        ctName := Find('name').AsString;
        ctCty := Find('country').AsString;
        jcoord := Find('coord') as TJSONObject;
        ctLon := jcoord.Find('lon').AsFloat;
        ctLat := jcoord.Find('lat').AsFloat;
        ctKoorIdx := CalcKoorIdx(ctLat, ctlon);
      end;
      pPOIs[j] := @POIs[j];
    end;
    j := jArr.Count;
  finally
    json.Free;
  end;
  Memo1.Lines.Add('Einlesen in ' + FormatDateTime('s.zzz"s"', now - t));
  //Sortieren nach Land,Name,ID
  mergesort(pPOIs, @CompSeq);
  Memo1.Lines.Add('+Sortieren in ' + FormatDateTime('s.zzz"s"', now - t));

  f := TFileStream.Create(ExtractFilePath(ParamStr(0)) + OutName,
    fmCreate or fmOpenWrite);
  j := j - 1;
  for i := 0 to j do
    f.Write(pPOIs[i]^, SizeOf(TCity));
  f.Free;
  Memo1.Lines.Add('+Wegschreiben in ' + FormatDateTime('s.zzz"s"', now - t));
  // Neu sortiert einlesen
  ReadPOIs(OutName);
end;

procedure TForm1.FillStringGrid;
var
  i, j: NativeInt;
begin
  StringGrid1.BeginUpdate;
  StringGrid1.RowCount := ResCnt + 1;

  for j := 0 to ResCnt - 1 do
    with tpCity(pPOIs[ResIdx[j]])^ do
    begin
      i := j + 1;
      StringGrid1.Cells[colCty, i] := ctCty;
      StringGrid1.Cells[colID, i] := IntToStr(ctid);
      StringGrid1.Cells[colLon, i] := FloatToStr(ctLon);
      StringGrid1.Cells[colLat, i] := FloatToStr(ctLat);
      StringGrid1.Cells[colName, i] := ctName;
    end;
  StringGrid1.EndUpdate;
end;

procedure TForm1.ReadPOIs(OutName: string);
var
  f: TFileStream;
  pPoi: tpCity;
  i, j: NativeInt;
begin
  f := TFileStream.Create(ExtractFilePath(ParamStr(0)) + OutName, fmOpenRead);
  j := f.Size;
  i := j div SizeOf(TCity);
  setlength(POIs, i);
  setlength(pPOIs, i);
  setlength(ResIdx, j);

  f.Read(POIs[0], j);
  f.Free;
  pPoi := @POIs[0];
  for j := 0 to High(POIs) do
  begin
    pPOIs[j] := pPoi;
    Inc(pPoi);
  end;
end;

procedure TForm1.OutPOI(POIidx: NativeInt);
begin
  with tpCity(pPOIs[POIidx])^ do
    //writeln(POIidx:8,ctID:8,ctCty:4,ctLon:12:6,ctLat:11:6,'  ',ctName);
    Memo1.Lines.add(Format('%8d %8d %4s %12.6f %12.6f %6d %s',
      [POIidx, ctID, ctCty, ctLon, ctLat, ctKoorIdx, ctName]));
end;

procedure TForm1.CheckCountry(s: string);
var
  i: NativeInt;
begin
  if length(s) <> 2 then
    EXIT;
  s := Uppercase(s);
  ResCnt := 0;
  for i := 0 to High(pPOIs) do
    if s = tpCity(pPOIs[i])^.ctCty then
    begin
      ResIdx[ResCnt] := i;
      Inc(ResCnt);
    end;
  FillStringGrid;
end;

procedure TForm1.CheckName(s: string);
var
  i: NativeInt;
begin
  s := lowercase(s);
  Rescnt := 0;
  for i := 0 to High(pPOIs) do
    if POS(s, lowercase(tpCity(pPOIs[i])^.ctName)) = 1 then
    begin
      ResIdx[Rescnt] := i;
      Inc(Rescnt);
    end;
  FillStringGrid;
end;

procedure TForm1.CheckCoor(coorIdx: NativeInt);
var
  i: NativeInt;
begin
  ResCnt := 0;
  for i := 0 to High(pPOIs) do
    if coorIdx = tpCity(pPOIs[i])^.ctKoorIdx then
    begin
      ResIdx[Rescnt] := i;
      Inc(Rescnt);
    end;
  FillStringGrid;
end;

procedure TForm1.Button1Click(Sender: TObject);
var
  T0, T1: int64;
  s: String;
  lon,lat : double;
  i: longint;
begin
  memo1.Clear;
  edName.Text := 'Berlin';
  edCountry.Text := 'DE';
  Latitude.Value:= 53.0;
  Longitude.Value:= 13.0;

  if FileExists(cOutName) then
  begin
    T0 := GetTickCount64;
    ReadPOIs(cOutName);
    T1 := GetTickCount64;
    Memo1.Lines.Add(Format(' Alles einlesen in %d ms', [T1 - T0]));
  end
  else
  begin
    if not (FileExists(cInName)) then
    begin
      Memo1.Lines.Add(Format('Keine passende Datei %s vorhanden', [cInName]));
      HALT(-2);
    end;
    T0 := GetTickCount64;
    ConvertJson(cInName, cOutName);
    T1 := GetTickCount64;
    Memo1.Lines.Add(Format(' Alles wegschreiben in %d ms', [T1 - T0]));
  end;
   SucheLand.Enabled := true;
   SucheOrt.Enabled := true;
   SucheKoor.Enabled:= true;
   Button1.Enabled:=false;
end;


procedure TForm1.FormShow(Sender: TObject);
begin
  setlength(CompArr, 3);
  CompArr[0] := @compCty;
  CompArr[1] := @CompKoorIdxID;
  CompArr[2] := @compName;
  memo1.Lines.Clear;
end;



procedure TForm1.SucheKoorClick(Sender: TObject);
var
  T1,T0:Int64;
  lat,lon:double;
  i : NativeInt;
begin
  T0 := GetTickCount64;
  lon := Longitude.value;
  lat := Latitude.Value;
  i := CalcKoorIdx(lat,lon);
  CheckCoor(i);
  T1 := GetTickCount64;
  Memo1.Lines.Add(Format(' Fand %d mit KoorIdx %d in %d ms', [ResCnt, i, T1 - T0]));
  application.ProcessMessages;
end;

procedure TForm1.SucheLandClick(Sender: TObject);
var
  T1,T0:Int64;
begin
  T0 := GetTickCount64;
  CheckCountry(edCountry.Text);
  T1 := GetTickCount64;
  Memo1.Lines.Add(Format(' Finde %d Orte in country %s in %d ms', [ResCnt, edCountry.Text, T1 - T0]));
  application.ProcessMessages;
end;

procedure TForm1.SucheOrtClick(Sender: TObject);
var
  T1,T0:Int64;
begin
  T0 := GetTickCount64;
  CheckName(edName.Text);
  T1 := GetTickCount64;
  Memo1.Lines.Add(Format(' Fand %d %s in %d ms', [ResCnt, edName.Text, T1 - T0]));
  application.ProcessMessages;
end;

begin
end.
{Einlesen in 2.531s
+Sortieren in 2.656s -> 125 ms
+Wegschreiben in 2.718s-> 62ms  ( 23 Mb ->370 mb/s Sata-SSD )
 Alles wegschreiben in 2734 ms
Es wurden 0 Übereinstimmungen gefunden
 Suche nach YX  in 16 ms
   10907  2148650   AU   152,483337   -24,916670 Springfield
   10908  6693094   AU   152,924881   -27,667570 Springfield Lakes
   23246  6154178   CA   -64,865494    44,633450 Springfield
  203996  4173892   US   -85,611320    30,153259 Springfield
  203997  4224162   US   -81,311501    32,372410 Springfield
  203998  4250542   US   -89,643707    39,801720 Springfield
  203999  4309329   US   -85,222183    37,685341 Springfield
  204000  4409896   US   -93,298241    37,215328 Springfield
  204001  4525353   US   -83,808823    39,924229 Springfield
  204002  4561407   US   -75,320190    39,930672 Springfield
  204003  4659557   US   -86,885002    36,509209 Springfield
  204004  4733920   US   -96,482758    31,659611 Springfield
  204005  4787117   US   -77,187202    38,789280 Springfield
  204006  4951788   US   -72,589813    42,101479 Springfield
  204007  5010917   US   -85,239159    42,326431 Springfield
  204008  5048394   US   -94,975822    44,238850 Springfield
  204009  5079488   US   -96,134460    41,081940 Springfield
  204010  5093030   US   -72,033417    43,495071 Springfield
  204011  5104952   US   -74,317230    40,704910 Springfield
  204012  5232077   US   -97,897293    42,854172 Springfield
  204013  5241423   US   -72,482307    43,298409 Springfield
  204014  5440032   US  -102,614357    37,408352 Springfield
  204015  5754005   US  -123,022034    44,046242 Springfield
Es wurden 23 Übereinstimmungen gefunden
 Suche nach Springfield  in 47 ms

 55  2861876   DE     7,000000    51,500000  48013 Land Nordrhein-Westfalen
194  6553078   DE     7,118200    51,034801  48013 Odenthal
197  6553073   DE     7,366500    51,033199  48013 Lindlar
200  2820086   DE     7,717500    51,568611  48013 Kreis Unna
208  2957818   DE     7,766670    51,583328  48013 Bönen
214  6553094   DE     7,616670    51,783298  48013 Ascheberg
1351  2947416   DE     7,216670    51,483330  48013 Bochum
}
